# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
"""Class for the Data faults and misconfigurations for AutoML."""
from typing import Dict, List, Optional, Any
import json
import os
import pkg_resources

from sklearn.preprocessing.imputation import Imputer
from azureml.automl.core._engineered_feature_names import FeatureTypeRecognizer
from azureml.automl.core.automl_run_context import AutoMLAbstractRunContext
from azureml.automl.core._vendor.automl.client.core.common.activity_logger import TelemetryActivityLogger
from azureml.automl.core.column_purpose_detection.types import StatsAndColumnPurposeType
from azureml.automl.core._vendor.automl.client.core.common.constants import AutoMLJson

from .featurization.data_transformer import DataTransformer


class VerifierResults:
    PASSED = "passed"
    ALERTED = "alerted"
    FIXED = "fixed"
    DONE = "done"
    STATUS = {PASSED, ALERTED, FIXED, DONE}


class VerifiedFaultsTypes:
    CLASS_BALANCEING = "_class_banlancing"
    TRAIN_TEST_SPLIT = "_train_test_split"
    MISSING_VALUE_IMPUTATION = "_missing_value"
    DETECT_HIGH_CARDINAL_FEATURES = "_detect_high_cardinal_feature"
    CROSS_VALIDATION = "_cross_validation"


class ValidParameterKeys:
    COLUMN_NAME = "column_name"
    MISSING_VALUES = "missing_values"
    IMPUTATION_TYPE = "imputation_type"
    DATASET = "dataset"
    ROW_COUNTS = "row_counts"
    PERCENTAGE = "percentage"
    NUMBER_OF_FOLDS = "number_of_folds"
    CONTENT_TYPE = "content_type"


class ValidParameterValues:
    MEAN = "mean"
    MODE = "mode"
    TRAIN = "train"
    TEST = "test"
    CATEGORICALHASH = "categorical_hash"
    TEXT = "text"


class VerifiedFault:
    def __init__(self,
                 type_name: str,
                 result: str,
                 parameters: Optional[List[Dict[str, Any]]],
                 is_child_run_specific: Optional[bool] = False):
        self.type = type_name
        self.result = result
        self.is_child_run_specific = is_child_run_specific
        if parameters is None:
            self.parameters = []   # type: List[Any]
        else:
            self.parameters = parameters

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": self.type,
            "result": self.result,
            "parameters": self.parameters
        }


class VerifiedResult:
    def __init__(self,
                 fault: VerifiedFault,
                 description: Dict[str, Any]):

        self.type = fault.type
        self.result = fault.result
        self.parameters = fault.parameters

        self.fr_type = description["name"]
        self.fr_result = self.add_details_to_result(description["states"][self.result])
        self.fr_parameters = []     # type: List[Dict[Any, str]]

        for param in fault.parameters:
            temp_dict = {}
            for each in param.keys():
                temp_dict[description["parameters"][each]["name"]] = param[each]
            self.fr_parameters.append(temp_dict)

    def add_details_to_result(self, message: str) -> str:
        if self.type == VerifiedFaultsTypes.DETECT_HIGH_CARDINAL_FEATURES and len(self.parameters) > 0:
            message = message.format(self.parameters[0][ValidParameterKeys.CONTENT_TYPE])
        return message

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": self.type,
            "result": self.result,
            "parameters": self.parameters,
            "friendly_type": self.fr_type,
            "friendly_result": self.fr_result,
            "friendly_parameters": self.fr_parameters
        }


class VerifierManager:
    PACKAGE_NAME = 'azureml.automl.core'
    REFERENCE_FILE_PATH = pkg_resources.resource_filename(PACKAGE_NAME, 'faults_verifier_message.json')
    LOCAL_RESULT_PATH = "verifier_results.json"

    def __init__(self):
        self.data_faults_names = []  # type: List[str]
        self.data_faults_dict = {}  # type: Dict[str, VerifiedFault]

    def _clear_obj(self) -> None:
        self.data_faults_names = []
        self.data_faults_dict = {}

    def has_fault_member(self, fault_name: str) -> bool:
        return fault_name in self.data_faults_dict.keys()

    def add_data_fault(self,
                       fault_name: str,
                       result: str,
                       parameters: Optional[List[Dict[str, Any]]] = None,
                       is_child_run_specific: Optional[bool] = False) -> None:
        self.data_faults_names.append(fault_name)
        self.data_faults_dict[fault_name] = VerifiedFault(fault_name,
                                                          result,
                                                          parameters,
                                                          is_child_run_specific)

    def add_data_fault_parameter(self, fault_name: str, parameters: Dict[str, str]) -> None:
        self.data_faults_dict[fault_name].parameters.append(parameters)

    def upload_to_file(self, run_context: AutoMLAbstractRunContext, ret_str: str, file_path: str) -> None:
        run_context.save_str_output(ret_str, file_path, overwrite_mode=True)

    def get_json_str(self) -> str:
        return json.dumps([self.data_faults_dict[m].to_dict() for m in self.data_faults_names])

    def get_elaborate_dicts(self, friendly_desc: Dict[str, Any]) -> List[Any]:

        dict_list = []

        for fault_name in self.data_faults_names:
            fault = self.data_faults_dict[fault_name]
            res_obj = VerifiedResult(fault, friendly_desc["types"][fault.type])
            dict_list.append(res_obj.to_dict())

        return dict_list

    def get_elaborate_json(self, run_context: AutoMLAbstractRunContext,
                           remote_path: str, logger: TelemetryActivityLogger) -> None:
        try:
            with open(self.REFERENCE_FILE_PATH, 'r') as f:
                friendly_desc = json.load(f)

            data_faults = self.get_elaborate_dicts(friendly_desc)

            ret_dict = {"schema_type": AutoMLJson.SCHEMA_TYPE_FAULT_VERIFIER,
                        "schema_version": friendly_desc['version'],
                        "faults": data_faults}

            ret_str = json.dumps(ret_dict)

            with run_context.get_run():
                try:
                    file_path = os.path.join(remote_path, self.LOCAL_RESULT_PATH)
                    self.upload_to_file(run_context, ret_str, file_path)
                    self._clear_obj()

                except Exception:
                    logger.warning("Failed uploading version files to artifact services.")

        except FileNotFoundError:
            logger.warning("Reference file, `faults_verifier_message.json` does not exist.")

    def to_json_file(self, path: str) -> None:
        with open(path, "w") as f:
            f.write(self.get_json_str())

    def to_json_file_run_id(self, run_id: str) -> None:
        self.to_json_file("{}_{}.json".format(run_id, "verifier_results"))

    def upload_to_artifact(self, automl_run_context: AutoMLAbstractRunContext) -> None:
        automl_run_context.save_str_output(
            self.get_json_str(), self.LOCAL_RESULT_PATH)

    def update_data_verifier_for_cv(self, number_of_folds: int) -> None:
        """
        Update VerifierManager object data fault with number of folds in Cross Validation.

        :param number_of_folds: Number of folds in cross validation.
        :return: None
        """
        self.add_data_fault(
            VerifiedFaultsTypes.CROSS_VALIDATION,
            result=VerifierResults.DONE,
            parameters=[{ValidParameterKeys.NUMBER_OF_FOLDS: str(number_of_folds)}]
        )

    def update_data_verifier_for_train_test_validation(self, train_row_count: int, test_row_count: int) -> None:
        """
        Update VerifierManager object data faults with the train/test sample ratio.

        :param train_row_count: Sample count in training set.
        :param test_row_count: Sample count in verification set.
        :return: None
        """
        self.add_data_fault(
            VerifiedFaultsTypes.TRAIN_TEST_SPLIT,
            result=VerifierResults.DONE,
            parameters=[
                {
                    ValidParameterKeys.DATASET: ValidParameterValues.TRAIN,
                    ValidParameterKeys.ROW_COUNTS: str(train_row_count),
                    ValidParameterKeys.PERCENTAGE: str(100. * train_row_count / (train_row_count + test_row_count))
                },
                {
                    ValidParameterKeys.DATASET: ValidParameterValues.TEST,
                    ValidParameterKeys.ROW_COUNTS: str(test_row_count),
                    ValidParameterKeys.PERCENTAGE: str(100. * test_row_count / (train_row_count + test_row_count))
                }
            ]
        )

    def update_data_verifier_from_transformer(self, data_transformer: DataTransformer) -> None:
        """
        Detect missing value imputation in DataTransformer and update VerifierManager object with fault.

        :param data_transformer: DataTransformer object.
        :return: None
        """
        column_names = data_transformer._columns_types_mapping.keys() \
            if data_transformer._columns_types_mapping else []
        if not data_transformer.transformer_and_mapper_list:
            return
        for transformer_mapper in data_transformer.transformer_and_mapper_list:
            for f in transformer_mapper.mapper.features:
                if isinstance(f[0], List):
                    col_name = f[0][0]
                else:
                    col_name = f[0]
                if col_name in column_names and isinstance(f[1][0], Imputer):
                    if not self.has_fault_member(VerifiedFaultsTypes.MISSING_VALUE_IMPUTATION):
                        self.add_data_fault(
                            VerifiedFaultsTypes.MISSING_VALUE_IMPUTATION, VerifierResults.FIXED)

                    self.add_data_fault_parameter(VerifiedFaultsTypes.MISSING_VALUE_IMPUTATION, {
                        ValidParameterKeys.COLUMN_NAME: str(col_name),
                        ValidParameterKeys.IMPUTATION_TYPE: f[1][0].strategy
                    })
        if not self.has_fault_member(VerifiedFaultsTypes.MISSING_VALUE_IMPUTATION):
            self.add_data_fault(VerifiedFaultsTypes.MISSING_VALUE_IMPUTATION, VerifierResults.PASSED)

    def update_data_verifier_for_text_class_validation(self, stats_and_column_purpose:
                                                       Optional[List[StatsAndColumnPurposeType]]) -> None:
        """
        Detect high cardinality features from DataTransformer and update VerifierManager object with fault.

        :param stats_and_column_purpose: Statistics and other info about columns from a data transformer.
        :return: None
        """
        if stats_and_column_purpose is None:
            stats_and_column_purpose = []

        for col in stats_and_column_purpose:
            column_purpose = col[1]
            if column_purpose == FeatureTypeRecognizer.CategoricalHash or column_purpose == FeatureTypeRecognizer.Text:
                content_type = ValidParameterValues.CATEGORICALHASH \
                    if column_purpose == FeatureTypeRecognizer.CategoricalHash else ValidParameterValues.TEXT
                if not self.has_fault_member(VerifiedFaultsTypes.DETECT_HIGH_CARDINAL_FEATURES):
                    self.add_data_fault(
                        VerifiedFaultsTypes.DETECT_HIGH_CARDINAL_FEATURES, VerifierResults.FIXED)
                self.add_data_fault_parameter(VerifiedFaultsTypes.DETECT_HIGH_CARDINAL_FEATURES, {
                    ValidParameterKeys.COLUMN_NAME: str(col[2]),
                    ValidParameterKeys.CONTENT_TYPE: content_type
                })
        if not self.has_fault_member(VerifiedFaultsTypes.DETECT_HIGH_CARDINAL_FEATURES):
            self.add_data_fault(VerifiedFaultsTypes.DETECT_HIGH_CARDINAL_FEATURES, VerifierResults.DONE)


def print_fault_misconfiguration_detection_results(Run: int) -> None:
    """Print the results of fault/misconfiguration detection results."""
    try:
        pass
    except Exception:
        print("Cannot find the fault/misconfiguration detection results file.")
